from PyQt5.QtWidgets import QVBoxLayout, QLabel, QHBoxLayout, QWizard, QWizardPage, QComboBox, QCheckBox
from PyQt5.QtCore import pyqtSignal
from yeastvision.models.proSeg.model import ProSeg
from yeastvision.models.budSeg.model import BudSeg
from yeastvision.parts.wizard_utils import InterpolationDialog, SimpleTextPage, ParameterInputPage

fiest_instructs = '''Frame Interpolation Enhanced Single-cell Tracking (FIEST),(FIEST):\n
Step 1: Interpolate images over specified intervals to enhance resolution\n
Step 2: Segment interpolation-enhanced images with proSeg\n
Step 3: Track resulting masks\n
Step 4 (optional): compute lineages\n
Step 5: De-interpolate masks, leaving only masks corresponding to original images\n'''

class DropDownCheckBoxPage(QWizardPage):
    completeChanged = pyqtSignal()
    def __init__(self, viable_channels, parent=None):
        super().__init__(parent)
        self.setTitle("Choose Channel And Lineages")

        # Create the label and combo box
        label = QLabel("Channel:")
        self.channel_select = QComboBox()
        self.channel_select.addItems([ts.name for ts in viable_channels])

        # Create the checkbox
        self.checkbox = QCheckBox("do_lineage")

        # Create a horizontal layout for the label and combo box
        h_layout = QHBoxLayout()
        h_layout.addWidget(label)
        h_layout.addWidget(self.channel_select)

        # Create the main layout and add the horizontal layout and checkbox to it
        layout = QVBoxLayout()
        layout.addLayout(h_layout)
        layout.addWidget(self.checkbox)
        self.setLayout(layout)

        # Register the combo box index field with automatic updates
        self.registerField("channelIndex*", self.channel_select, "currentIndex", self.channel_select.currentIndexChanged)
        self.registerField("doLineage*", self.checkbox)

        # Connect signals to check for completeness
        self.channel_select.currentIndexChanged.connect(self.completeChanged.emit)
        self.checkbox.stateChanged.connect(self.completeChanged.emit)

    def isComplete(self):
        # Page is complete if a valid channel is selected (index >= 0)
        return self.channel_select.currentIndex() >= 0

    def getData(self):
        return {
            "channelIndex": self.field("channelIndex"),
            "doLineage": self.field("doLineage")
        }

class ParameterInputPageFiestAsexual(ParameterInputPage):
    def __init__(self, params_dict, model_weights, modelName, channels, parent=None, includeTime=True):
        super().__init__(params_dict, model_weights, modelName, channels, parent=parent, includeTime=includeTime)
    
    def nextId(self):
        if self.field("doLineage"):
            return 5  # ID of the additional DisplaySelectionPage
        else:
            return -1  # Complete the wizard
    
class AdditionalParameterInputPage(ParameterInputPageFiestAsexual):
    def __init__(self, params_dict, model_weights, modelName, channels, parent=None, includeTime=True):
        super().__init__(params_dict, model_weights, modelName, channels, parent=parent, includeTime=includeTime)
    
    def nextId(self):
        return -1
    
class FiestWizard(QWizard):
    def __init__(self, parent, channels, proSeg_weights, budSeg_weights):
        super().__init__(parent)
        
        self.setWindowTitle("Frame Interpolation Enhanced Segmentation Tracking")

        self.setPage(1, SimpleTextPage(fiest_instructs))
        self.setPage(2, DropDownCheckBoxPage(channels, parent=self))
        self.setPage(3, InterpolationDialog(channels, parent=self))
        self.setPage(4, ParameterInputPage(ProSeg.hyperparams, proSeg_weights, "proSeg", channels, parent=self, 
                                           includeTime=False))
        self.setPage(5, AdditionalParameterInputPage(BudSeg.hyperparams, budSeg_weights, "budSeg", channels, parent=self, 
                                                     includeTime=False))

        self.non_model_param_ids = [2,3]
        self.cyto_seg_id = 4
        self.bud_seg_id = 5
    
    def getData(self):
        data = {}
        for page_id in self.non_model_param_ids:
            page = self.page(page_id)
            page_data = page.getData()
            data.update(page_data)
        proSeg_page = self.page(self.cyto_seg_id)
        data["proSeg"] = proSeg_page.getData()
        if self.field("doLineage"):
            bud_page = self.page(self.bud_seg_id)
            data["budSeg"] = bud_page.getData()
        return data


import logging
import numpy as np
from skimage.morphology import thin
import statistics
from tqdm import tqdm

logger = logging.getLogger(__name__)

def track_to_cell(objs:np.ndarray, cells:np.ndarray)->np.ndarray:
    '''
    Renumbers labels that are associated with tracked cell labels to the same index as their corresponding cells. 
    Useful for ensuring that vacuoles or nucleus have the same index as their corresponding cell label. The algorithm is simple
    and assigns an index to label based on the index it overlaps the most with in cells
    
    Args: 
        objs (np.ndarray): a timeseries of labeled cell features that will be rebnumbered based on cells
        cells (np.ndarray): should be same shape as objs. Contains tracked cells
    
    Returns:
        np.ndarray: a 3D array where the pixels are all times are renumbered based on the max overlapping region in cells at the same timepoint '''
    tracked_mask = np.zeros_like(objs)
    im_idx = 0 
    for obj_mask, cell_mask in zip(objs, cells):
        for obj_num in np.unique(obj_mask):
            if obj_num > 0:
                vals, counts  = np.unique(cell_mask[obj_mask==obj_num], return_counts=True)

                if np.any(vals==0):
                    vals, counts = vals[1:], counts[1:]

                tracked_mask[im_idx][obj_mask==obj_num] = vals[np.argmax(counts)]
        im_idx += 1
    return tracked_mask


def extract_cell_frames(ims: np.ndarray) -> int:
    """
    Determines the index of the first frame after cells disappear in a series of images.

    This function scans through a sequence of 2D images and returns the index of the first 
    frame where the cells have disappeared. It assumes that cells are present when any 
    pixel value is greater than 0 and that the absence of cells is indicated by all pixel 
    values being 0.

    Args:
        ims (np.ndarray): A 3D numpy array of shape `(num_frames, height, width)` representing 
                          a sequence of 2D images.

    Returns:
        int: The index of the first frame after cells disappear. If cells never disappear,
             it returns the index of the last frame.
    """
    found_cells = False
    for i, im in enumerate(ims):
        if not found_cells:
            found_cells = np.any(im > 0)
        if found_cells:
            if np.all(im == 0):
                break
    return i + 1

def track_proliferating(ims:np.ndarray)->np.ndarray:
    '''
    Tracking algorithm for asexually proliferating yeasts. Utilizes overlaps and assumption of
    limited yeast motility. For movies with rapibly expanding colonies or yeasts that move a lot,
    see yeastvision.track.fiest module
    
    Args:
        ims (np.ndarray): A 3d array of cell masks corresponding to proliferating cells. The first dimension should specify timepoints
    
    Returns:
        np.ndarray: A tracked, 3d array of the same shape and type as ims
    '''
    stop_i = extract_cell_frames(ims)
    ims_template = np.zeros_like(ims, dtype=np.uint16)
    ims = ims[:stop_i]
    logger.info("--TRACKING: LOOP 1/4")
    IS1 = ims[0,:,:]
    IblankG = np.zeros(IS1.shape, dtype="uint16") # allocate
    masks = np.zeros((IS1.shape[0], IS1.shape[1], ims.shape[0])) # allocate
    masks[:,:,0]= IS1   

    for it0 in tqdm(range(1,ims.shape[0])):
        m0b = ims[it0]
        IS2 = m0b.copy()
        IS2C = IS2 
        IS1B = IS1 
        IS1B= (IS1 > 0.5).astype(np.uint16) # binarize # plt.imshow(IS1B)
        
        IS3 = np.multiply(IS1B,IS2) # plt.imshow(IS3)
        
        tr_cells=(np.unique(IS1.astype(np.int0)))
        tr_cells=tr_cells[1:len(tr_cells)]
        gap_cells=(np.unique(IblankG.astype(np.int0)))
        gap_cells=gap_cells[1:len(gap_cells)]
        
        # gap_cells=np.array((10,122)).astype(np.uint64) 
        # cells_tr =  cells_tr + gap_cells
        if gap_cells.size==0 :
            cells_tr = tr_cells
        else:
        # cells_tr = np.array([tr_cells, gap_cells]) # !
        # cells_tr =  (tr_cells) + (gap_cells)
            cells_tr = np.concatenate((tr_cells,gap_cells),axis=0) # right concatenation
            cells_tr = np.sort(cells_tr)
            
        Iblank0=np.zeros(IS1.shape, dtype="uint16") # plt.imshow(Iblank0)
    
        for it1 in cells_tr: # ascending, default
            # it1=1
            IS5=(IS1==it1) # plt.imshow(IS5) # sum(sum(IS5))
            IS5=IS5.astype(np.uint16) # plt.imshow(IS5)
            IS56 = thin(IS5,1) # plt.imshow(IS56)  #  sum(sum(IS56))
            IS6A = np.multiply(IS56,IS3) # plt.imshow(IS6A)  #   sum(sum(IS6A))
        
            if sum(sum(IS5))==0 :
                IS5=(IblankG==it1)
                IS6A = np.multiply(IS56,IS2C)
                IblankG[IblankG==it1]=0
                # IblankG=np.where(IblankG==it1,IblankG,0)
            
            if sum(sum(IS6A))>0 :
                IS2ind=(statistics.mode(IS6A[np.nonzero(IS6A)])) # nonzero gives the indexes
                Iblank0[IS2==IS2ind]=it1
                IS3[IS3==IS2ind]=0
                IS2C[IS2==IS2ind]=0
                


        bl_cells = np.unique(Iblank0) # plt.imshow(Iblank0)
        bl_cells = bl_cells[1:len(bl_cells)]
        seg_gap = np.setdiff1d(tr_cells,bl_cells) 
        
                
        if seg_gap.size>0 :
            for itG in seg_gap:
                IblankG[IS1==itG]=itG
        
        
        Iblank0B = np.copy(Iblank0)      # plt.imshow(Iblank0)
        Iblank0B[np.nonzero(Iblank0B)] = 1  
        Iblank0B = (Iblank0B < 0.5).astype(np.uint16) # plt.imshow(Iblank0B) # invert Iblank0B
        
        ISB = np.multiply(IS2,Iblank0B) # plt.imshow(ISB)
        
        newcells=np.unique(ISB)
        newcells = newcells[1:len(newcells)]
        Iblank=Iblank0  # plt.imshow(Iblank)
        A=1
        
        if newcells.size>0 :
            for it2 in newcells :
                Iblank[IS2==it2]=max(cells_tr)+A
                A=A+1 
                
        masks[:,:,it0]=Iblank # plt.imshow(masks[:,:,it0-1])
        IS1=masks[:,:,it0] # plt.imshow(IS1)


    ccell2=np.unique(masks[:,:,len(masks[0][0])-1])
    ccell2 = ccell2[1:len(ccell2)] # remove zero since its background
    Mask2 = np.zeros((masks.shape[0], masks.shape[1], len(masks[0][0])))

    logger.info("TRACKING: LOOP 2/4")
    for itt4 in tqdm(range(0,len(masks[0][0]))) : # masks
        mask3=np.zeros((masks.shape[0], masks.shape[1]))
                
                
        for itt3 in range(0,len(ccell2)) : # % itt3=23
            pix2 = np.nonzero(masks[:,:,itt4]==ccell2[itt3])
            mask3[pix2]=itt3 + 1 # zero correction    
                
        Mask2[:,:,itt4] = mask3 

    no_obj=len(ccell2)
    all_ob=np.zeros((no_obj,len(Mask2[0][0])))
    logger.info("TRACKING: LOOP 3/4")
    for ccell in tqdm(range(0,no_obj)): # ccell=0
            
            for itt in range(0,len(masks[0][0])): # itt = 0
                Mas=Mask2[:,:,itt]==ccell + 1 # % imagesc(Mask2)
                pix=sum(sum(Mas))
                all_ob[ccell,itt]=pix #  figure;imagesc(all_ob)

    cell_exists=np.zeros((1,len(all_ob[:,0]))).astype("uint16")
    logger.info("TRACKING: LOOP 4/4")
    for itt2 in tqdm(range(0,len(all_ob[:,0]))) : # itt2 = 20
        etime = sum(all_ob[itt2,:]==0) #
        if etime == 0 :
            cell_exists[0,itt2]=1
        else:
            cell_exists[0,itt2]=np.max(np.nonzero(all_ob[itt2,:]==0)) + 1

    Mask2 = np.array([Mask2[:,:,i] for i in range(Mask2.shape[-1])], dtype = np.uint16)
    ims_template[:stop_i] = Mask2
    return ims_template



    




from os.path import join
import numpy as np
from os.path import join
from skimage.exposure import equalize_adapthist as ea
from yeastvision.utils import *
from yeastvision.models.utils import MODEL_DIR, patchify_for_train
import tqdm

class Model():
    hyperparams = {"Threshold":0.50}
    loss = None
    trainparams = {"learning_rate": 0.001, "n_epochs": 100}
    types = [None]
    prefix = ""

    def __init__(self, params, weights):
        self.params = params 
        self.weights = weights
    
    def getProbIm(self,im):
        return im
    
    def getMask(self, probIm):
        return (probIm>self.params["Threshold"]).astype(np.uint8)
    
    def preprocess(self, im):
        return ea(im).astype(np.float32)
    
    def preprocess_masks(self, mask):
        return (mask>0).astype(np.float32)
    
    def name(self):
        pass
    
    def train(self, ims, masks, params):
        batchsize = 8
        weightPath = join(params["dir"], "models", params["model_name"])+".h5"
        ims = [self.preprocess(im) for im in ims]
        masks = [self.preprocess_masks(mask) for mask in masks]
        ims = patchify_for_train(ims)
        masks = patchify_for_train(masks)
        ims = np.array(ims)
        masks = np.array(masks)
        model = self.model
        model.fit(ims, masks,
            batch_size=batchsize,
            epochs=int(params["n_epochs"]))

        # Save the trained model (optional)
        model.save(weightPath)

    @classmethod
    def run(cls, ims, params, weights):
        params = params if params else cls.hyperparams
        model = cls(params, weights)
        probIms, masks = [], []
        for im in tqdm(ims):
            probIm = model.getProbIm(im)
            probIms.append(probIm)
            masks.append(model.getMask(probIm))
        del model.model
        del model


        masks, probIms =  np.array(masks, dtype = np.uint16), (np.array(probIms)*255).astype(np.uint8)
        return masks, probIms






from yeastvision.models.cp import CustomCPWrapper

class FlouSeg(CustomCPWrapper):
    hyperparams  = { 
    "mean_diameter":30.0,
    "flow_threshold":0.9, 
    "cell_probability_threshold": 0.5}
    types = [float, float,float]

    def __init__(self, params, weights):
        super().__init__(params, weights)


from setuptools import setup
import os


# Function to automatically find all files in a specific directory
def find_files(directory):
    for path, directories, files in os.walk(directory):
        for file in files:
            print(file)
            yield os.path.join(path, file)

requires = [
"numpy==1.24",
"cellpose==2.1.0",
"charset-normalizer==3.3.2",
"matplotlib",
"munkres",
"opencv_python_headless",
"pandas==1.5.0",
"patchify==0.2.3",
"PyQt5==5.15.10",
"pyqtgraph==0.13.0",
"scikit_image",
"scikit_learn",
"scipy",
"torch==1.12.0",
"tqdm",
"trackpy",
"torchvision==0.13.1",
"chardet==5.2.0",
"memory-profiler",
"QSwitchControl"]

try:
    import torch
    a = torch.ones(2, 3)
    major_version, minor_version, _ = torch.__version__.split(".")
    if major_version == "2" or int(minor_version) >= 12:
        requires.remove("torch==1.12.0")
        requires.remove("torchvision==0.13.1")
except:
    pass


github_only_text  = '''
## Feature: Enhance time-series resolution with generative AI
<img src="https://raw.githubusercontent.com/berkyalcinkaya/yeastvision/main/yeastvision/docs/figs/interp.gif?raw=True" width="2000" height="600" />
'''

with open("README.md", "r") as fh:
    long_description = fh.read()
    # remove github only text from long description
    long_description = long_description.replace(github_only_text, "")

packages = [
            "yeastvision","yeastvision.data",  "yeastvision.plot", 
            "yeastvision.track", "yeastvision.track.fiest", "yeastvision.models", 
            "yeastvision.ims", "yeastvision.ims.rife_model", 
            "yeastvision.docs", "yeastvision.docs.figs",
            "yeastvision.ims.rife_model.pytorch_msssim", "yeastvision.parts", 
            "yeastvision.flou", "yeastvision.disk", "yeastvision.models.proSeg", 
            "yeastvision.models.budSeg", "yeastvision.models.budSeg",
            "yeastvision.models.matSeg", "yeastvision.models.spoSeg", 
            "yeastvision.models.flouSeg"
        ]

setup(
    name = "yeastvision",
    version = "0.1.70",
    description = "Deep learning-enabled image analysis of the full yeast life cycle",
    author = "Berk Yalcinkaya",
    url = "https://github.com/berkyalcinkaya/yeastvision",
    long_description=long_description,
    long_description_content_type='text/markdown',
    author_email="berkyalcinkaya55@gmail.com",
    license = "BSD",
    packages = packages,
    install_requires = requires,
    include_package_data=True,
    classifiers=(
        "Programming Language :: Python :: 3.10",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
    ),
    entry_points = {
        'console_scripts': [
          'yeastvision = yeastvision.__main__:main']
       }
)



import numpy as np
from skimage.morphology import disk, dilation, opening
from scipy.ndimage import zoom
from yeastvision.models.proSeg.model import ProSeg
from yeastvision.models.budSeg.model import BudSeg
from yeastvision.models.matSeg.model import MatSeg
from yeastvision.models.spoSeg.model import SpoSeg
from yeastvision.models.utils import produce_weight_path
import os
import numpy as np
import cv2

import os
import numpy as np
import cv2

def save_images_to_directory(images, save_dir, timepoint_dim=0, file_prefix="img"):
    """
    Saves a set of NumPy arrays representing images to a directory with a naming convention
    that ensures the images are sorted alphabetically by timepoint.

    If any image is None or an empty array, a zero array of the same shape as the previous timepoint is saved.
    If the first image is None or empty, a default 100x100 zero array is saved.

    Parameters:
    - images: A list of 2D NumPy arrays or a 3D NumPy array where one dimension represents timepoints.
    - save_dir: The directory where the images will be saved.
    - timepoint_dim: The dimension representing timepoints (0 for first dimension, -1 for last dimension).
    - file_prefix: A prefix for the filenames (default: "img").
    
    Output:
    - Saved images in the directory with filenames formatted to ensure correct alphabetic sorting.
    """
    if not os.path.exists(save_dir):
        os.makedirs(save_dir)

    # Handle the case where a list of arrays is provided
    if isinstance(images, list):
        num_timepoints = len(images)
        img_list = images
    else:
        # If a 3D numpy array is provided, adjust the timepoints according to the timepoint_dim
        num_timepoints = images.shape[timepoint_dim]
        if timepoint_dim == 0:
            img_list = [images[t, :, :] for t in range(num_timepoints)]
        elif timepoint_dim == -1:
            img_list = [images[:, :, t] for t in range(num_timepoints)]
        else:
            raise ValueError("Invalid timepoint_dim. Supported values are 0 or -1.")

    # Initialize the previous image for fallback in case of None or empty images
    prev_img = None

    for t, img in enumerate(img_list):
        if img is None or img.size == 0 or not img.shape:
            if prev_img is not None:
                # Use a zero array of the same shape as the previous timepoint
                img = np.zeros_like(prev_img)
            else:
                # For the first timepoint, use a default 100x100 zero array
                img = np.zeros((100, 100), dtype=np.uint8)
        
        # Update the previous image
        prev_img = img

        # Zero-padding the file name to ensure correct sorting
        filename = f"{file_prefix}_{str(t).zfill(5)}.png"  # e.g., img_00001.png
        filepath = os.path.join(save_dir, filename)

        # Save the image using OpenCV
        cv2.imwrite(filepath, img)

# Example usage:
# Assuming 'images' is a 3D numpy array with timepoints along dimension 0
# save_images_to_directory(images, "output_directory", timepoint_dim=0, file_prefix="timepoint")



def synchronize_indices(movie1: np.ndarray, movie2: np.ndarray) -> np.ndarray:
    """
    Synchronizes the indices of labeled objects in movie2 with those in movie1 based on pixel overlap.

    Parameters:
    movie1 (np.ndarray): A 3D array (timepoints x rows x cols) with labeled objects at each timepoint.
    movie2 (np.ndarray): A 3D array (timepoints x rows x cols) with labeled objects at each timepoint.

    Returns:
    np.ndarray: The updated movie2 with synchronized indices.
    """
    # Ensure movie1 and movie2 have the same dimensions
    if movie1.shape != movie2.shape:
        raise ValueError("movie1 and movie2 must have the same dimensions")

    # Copy movie2 to avoid modifying the original
    synced_movie2 = movie2.copy()

    # Iterate over each timepoint
    for t in range(movie1.shape[0]):
        # Get the labeled frames for movie1 and movie2 at time t
        frame1 = movie1[t]
        frame2 = movie2[t]

        # Get unique object labels from both frames, ignoring background (label 0)
        labels1 = np.unique(frame1[frame1 > 0])
        labels2 = np.unique(frame2[frame2 > 0])

        # Iterate over objects in movie2
        for label2 in labels2:
            # Find the overlap with any object in movie1
            overlap_mask = (frame2 == label2) & (frame1 > 0)

            if np.any(overlap_mask):
                # Find the most common label in movie1 for the overlapping region
                overlapping_labels = frame1[overlap_mask]
                most_common_label = np.bincount(overlapping_labels).argmax()

                # Set the corresponding pixels in movie2 to the most common label in movie1
                synced_movie2[t][frame2 == label2] = most_common_label

    return synced_movie2


def fill_empty_arrays(array_list):
    # Find the shape of the first non-empty array
    reference_shape = None
    for arr in array_list:
        if arr.size != 0:  # Check if the array is not empty
            reference_shape = arr.shape
            break
    
    # Replace empty arrays with zero arrays of reference_shape
    result_list = []
    for arr in array_list:
        if arr.size == 0:  # Empty array
            result_list.append(np.zeros(reference_shape))
        else:
            result_list.append(arr)
    
    return result_list

def resize_image(image, target_shape):
    zoom_factors = [n / float(o) for n, o in zip(target_shape, image.shape)]
    return zoom(image, zoom_factors, order=0)

def extend_seg_output(to_segment, output, start, stop):
    correct_len = []
    for i in range(len(output)):
        template = np.zeros_like(to_segment, dtype=np.uint16)
        template[start:stop] = output[i]
        correct_len.append(template)
    return correct_len


def _get_proSeg(proSeg_params, proSeg_weights)->ProSeg: 
    if not proSeg_params:
        proSeg_params = ProSeg.hyperparams
    if not proSeg_weights:
        proSeg_weights = produce_weight_path("proSeg", "proSeg")
    return ProSeg(proSeg_params, proSeg_weights), proSeg_params, proSeg_weights

def _get_matSeg(matSeg_params, matSeg_weights)->MatSeg: 
    if not matSeg_params:
        matSeg_params = MatSeg.hyperparams
    if not matSeg_weights:
        matSeg_weights = produce_weight_path("matSeg", "matSeg")
    return MatSeg(matSeg_params, matSeg_weights), matSeg_params, matSeg_weights

def _get_spoSeg(spoSeg_params, spoSeg_weights)->SpoSeg: 
    if not spoSeg_params:
        spoSeg_params = SpoSeg.hyperparams
    if not spoSeg_weights:
        spoSeg_weights = produce_weight_path("spoSeg", "spoSeg")
    return SpoSeg(spoSeg_params, spoSeg_weights), spoSeg_params, spoSeg_weights

def _get_budSeg(budSeg_params, budSeg_weights)->BudSeg: 
    if not budSeg_params:
        budSeg_params = BudSeg.hyperparams
    if not budSeg_weights:
        budSeg_weights = produce_weight_path("budSeg", "budSeg")
    return BudSeg(budSeg_params, budSeg_weights), budSeg_params, budSeg_weights


def remove_artif(I2A,disk_size): # I2A = IS2 % disk radius is 3 for ~500x~1000, 6 for larger images
    # we need a function to define the disk size base in the average cell size
        I2AA=np.copy(I2A) #   plt.imshow(IS2)
        # Applying logical operation and morphological opening
        I2A1 = binar(I2A);#binar(I2A) plt.imshow(I2A1)     plt.imshow(I2A)
    

        # Create a disk-shaped structuring element with radius 3
        selem = disk(disk_size)
        # Perform morphological opening
        I2B = opening(I2A1, selem)

        
        # Morphological dilation   plt.imshow(I2B)
        I2C = dilation(I2B, disk(disk_size))  # Adjust the disk size as needed


        I3 = I2AA * I2C # plt.imshow(I3)

        # Extract unique objects
        objs = np.unique(I3)
        objs = objs[1:len(objs)]
        
        # Initialize an image of zeros with the same size as I2A
        I4 = np.uint16(np.zeros((I3.shape[0], I3.shape[1])))
        # Mapping the original image values where they match the unique objects
        AZ=1
        for obj in objs:
            I4[I2A == obj] = AZ
            AZ=AZ+1
        
        return I4

#Helper Functions

def OAM_23121_tp3(M, cel, no_obj1, A):
    tp3 = M.copy()
    tp3[tp3 == cel] = no_obj1 + A
    return tp3

def binar(IS1):
    # Copy IS1 to IS1B
    IS1B = np.copy(IS1)
    
    # Convert non-zero elements to 1
    IS1B[IS1 != 0] = 1

    
    return IS1B

def cal_allob2(ccel, TETC, rang): 
     # Initialize the all_obj array with zeros
     all_obj = np.zeros((ccel, len(TETC)))

     for iv in range(ccel):  # Adjusted to 1-based index
         for its in rang:
             if TETC[its] is not None: #and np.sum(TETC[0][its]) > 0:  # Check if the array is not None and not empty
                 all_obj[iv, its] = np.sum(TETC[its] == iv + 1)  # Adjusted for 1-based index logic
             else:
                 all_obj[iv, its] = -1

     return all_obj

def cal_celldata(all_obj, ccel):
    cell_data = np.zeros((ccel, 5))

    for iv in range(ccel):
        first_occurrence = np.argmax(all_obj[iv, :] > 0)
        last_occurrence = len(all_obj[iv, :]) - np.argmax((all_obj[iv, :][::-1] > 0)) - 1
        
        cell_data[iv, 0] = first_occurrence  # 1st occurrence
        cell_data[iv, 1] = last_occurrence   # Last occurrence
        
    for iv in range(ccel):
        cell_data[iv, 2] = cell_data[iv, 1] - cell_data[iv, 0] + 1  # Times the cell appears
        aa1 = all_obj[iv, :]
        aa2 = aa1[int(cell_data[iv, 0]):int(cell_data[iv, 1]) + 1]
        aa3 = np.where(aa2 == 0)[0]
        cell_data[iv, 3] = len(aa3)  # Number of times it disappears between 1st and last occurrence
        cell_data[iv, 4] = (cell_data[iv, 3] * 100) / cell_data[iv, 2]  # Percentage of times the cell disappears

    return cell_data


def cal_allob(ccel, TETC, rang):
    # Initialize the all_obj array with zeros
    all_obj = np.zeros((ccel, len(TETC[0])))

    for iv in range(0, ccel):  # Adjusted to 1-based index
        for its in rang:
            if TETC[0][its] is not None:  # Check if the array is not None and not empty
                all_obj[iv, its] = np.sum(TETC[0][its] == iv + 1)  # Adjusted for 1-based index logic
            else:
                all_obj[iv, its] = -1

    return all_obj

# Removing artifacts - cells that appear once and cells that disappear thresh % of the time or more
def cal_allob1(ccel, TETC, rang):
    # Initialize the all_obj array with zeros
    all_obj = np.zeros((ccel, len(TETC[1])))

    for iv in range(ccel):  # Adjusted to 1-based index
        for its in rang:
            if TETC[0][its] is not None: #and np.sum(TETC[0][its]) > 0:  # Check if the array is not None and not empty
                all_obj[iv, its] = np.sum(TETC[0][its] == iv + 1)  # Adjusted for 1-based index logic
            else:
                all_obj[iv, its] = -1

    return all_obj

def replace_none_with_empty_array(data):
    if isinstance(data, list):
        return [replace_none_with_empty_array(item) for item in data]
    elif data is None:
        return np.array([])
    else:
        return data
    



from PyQt5.QtCore import Qt
from PyQt5 import QtWidgets, QtGui, QtCore
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt

class TableModel(QtCore.QAbstractTableModel):
    def __init__(self, data, main):
        super(TableModel, self).__init__()
        self.main = main
        self._data = data
        self.columns = {"cell":0, "birth":1, "death": 2, "mother":3, "confidence":4}
        self.hasLineages = "mother" in self._data.columns
        self.confidenceGradColors = plt.get_cmap("Greens", 12)
        if self.hasLineages:
            self.findPotentialErrors()

    def data(self, index, role):
        if index.isValid():
            if role == Qt.DisplayRole or role == Qt.EditRole:
                value = self._data.iloc[index.row(), index.column()]
                return str(value)
            
        if role == Qt.DisplayRole:
            value = (self._data.iloc[index.row(), index.column()]).copy()
            return str(value)
        
        if role == Qt.BackgroundRole:
            if self.hasLineages:
                if (index.row(), index.column()) in self.turnRed:
                    return QtGui.QColor('red')
                if index.column() == self.columns["confidence"]:
                    value = (self._data.iloc[index.row(), index.column()]).copy()
                    value = float(value)
                    value = round(value, 1)*10
                    r,g,b,a = self.confidenceGradColors(int(value)+1)
                    return QtGui.QColor(int(r*255),int(g*255),int(b*255),int(a*125))

    def rowCount(self, index):
        return self._data.shape[0]

    def columnCount(self, index):
        return self._data.shape[1]

    def headerData(self, section, orientation, role):
        # section is the index of the column/row.
        if role == Qt.DisplayRole:
            if orientation == Qt.Horizontal:
                return str(self._data.columns[section])

            if orientation == Qt.Vertical:
                return str(self._data.index[section])
    
    def flags(self, index):
        if index.column() ==  self.columns["mother"]:
            return Qt.ItemIsSelectable|Qt.ItemIsEnabled|Qt.ItemIsEditable
        else:
            return Qt.ItemIsSelectable|Qt.ItemIsEnabled
    
    def setData(self, index, value, role):
        if role == Qt.EditRole:
            self._data.iloc[index.row(),index.column()] = value
            self.main.saveData()
            return True
            
    def findPotentialErrors(self):
        groups = self._data.groupby(["mother", "birth"])
        result = groups.filter(lambda x: len(x) > 1)
        idxs = result.index.tolist()
        self.turnRed = list(zip(idxs, [self.columns["mother"]]*len(idxs)))
    
    def handleselect(self, new, previous):
        if previous.column() == -1:
            return
        


        if (new.column() == self.columns["cell"] or new.column() == self.columns["mother"]):
            value = self._data.iloc[new.row(), new.column()]
            if value == np.NaN:
                return
            birth = int(self._data.iloc[value-1, self.columns["birth"]])
            self.main.tIndex = birth
            self.main.selectCellFromNum(value)

        
        if (new.column() in [self.columns["birth"], self.columns["death"]]):
            self.main.tIndex = int(self._data.iloc[new.row(), new.column()])
            self.main.drawIm()
            self.main.drawMask()

        

        



from typing import List, Tuple
import cv2
from torch.nn import functional as F
import torch
import numpy as np
from yeastvision.ims.rife_model.pytorch_msssim import ssim_matlab
from queue import Queue, Empty
from .rife_model.RIFE import Model
from yeastvision.ims import rife_model
from yeastvision.data.ims import InterpolatedChannel
import copy
import os

RIFE_DIR = rife_model.__path__[0]
RIFE_WEIGHTS_NAME = "flownet.pkl"
RIFE_WEIGHTS_PATH = os.path.join(RIFE_DIR, RIFE_WEIGHTS_NAME)


def get_interp_labels(original_len: int, intervals: List[dict])->List[bool]:
    '''Produces a list of boolean flags that is parrallel to the number of interpolated images, where
    True corresponds to all frame that are interpolated and false corresponds to all real images
    
    Parameters
    - original_len (int): length of the original movie without interpolation
    - intervals (list of dictionaries): the interpolation intervals that were input to interpolate_intervals
    
    Returns
    - interp_labels (list of bools)'''
    
    interp_labels = [False for i in range(original_len)]
    interp_labels = InterpolatedChannel.insert_interpolated_values(interp_labels, intervals, True)
    return interp_labels
    
def deinterpolate(interpolated_ims:np.ndarray[np.uint8], interp_labels: List[bool])->List[np.ndarray[np.uint8]]:
    '''Isolates the real images in interolated_ims and returns them as a new list.
    
    Parameters
    - interpolated_ims (uint8 np.ndaray): output of RIFE interpolation
    - interp_labels (list): a list parrallel to the length of interpolate ims that specifies the location of interpolated frames with True values, False for real-images '''
    return [im for i, im in enumerate(interpolated_ims) if not interp_labels[i]]


def get_interpolated_length(t: int, intervals: List[dict])->int:
    """
    Calculate the number of frames after interpolation.

    Parameters:
    t (int): Original number of frames in the movie.
    intervals (list): List of interval dictionaries with 'start', 'stop', and 'interp' keys.

    Returns:
    int: Total number of frames after interpolation.
    """
    total_frames = t

    for interval in intervals:
        start = interval['start']
        stop = interval['stop']
        interp_factor_log2 = int(interval['interp'])
        num_frames = stop - start
        additional_frames = (num_frames) * (2**interp_factor_log2 - 1)
        total_frames += additional_frames
    return int(total_frames)

def rife_weights_loaded():
    return os.path.exists(RIFE_WEIGHTS_PATH)

def copy_and_normalize(im):
    new_im = copy.deepcopy(im)
    return cv2.normalize(new_im, None, 0, 255, cv2.NORM_MINMAX, dtype=cv2.CV_8U)

def load_model()->Tuple[Model, str]:
    '''
    Instantiates the RIFE model class and determines device.'''
    device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
    torch.set_grad_enabled(False)
    if torch.cuda.is_available():
        torch.backends.cudnn.enabled = True
        torch.backends.cudnn.benchmark = True
    
    model = Model()
    model.load_model(RIFE_DIR, -1)
    model.eval()
    model.device()
    return model, device

def interpolate_intervals(ims: np.ndarray, intervals: List[dict])->Tuple[np.ndarray, List[dict]]:
    '''
    Interpolates ims over an arbitrary number of intervals
    
    Parameters
    ims (np.ndarray): an (n x h x w) array with n images of h x w dimensions to be interpolated
    interp_intervals (lists of dict): list specifying interpolation levels and intervals, each entry
                                                of the list should be a dictionary with the keys (start, stop, interp)
                                                where interp is an integer value in [1,2,3,4]. If none, no interpolation is
                                                performed
                                                
    Returns
    interpolated (np.ndarray): resulting images (uint8) that have been interpolated over the specified intervals. Has same height and width dimensions 
                                as original
    new_intervals (List[dict]): a list of dictionaries, each with the keys start, stop, and interp, that specify where interpolated frames exist in 
    'interpolated'
    '''
    model, device = load_model()
    single_im_shape = ims[0].shape
    new_length = get_interpolated_length(len(ims), intervals)
    #print(intervals)
    #print((new_length, single_im_shape[0], single_im_shape[1]))
    template = np.zeros((new_length, single_im_shape[0], single_im_shape[1]), dtype=np.uint8)

    template_pos = 0  # This keeps track of where we are in the new template
    original_pos = 0
    interval_index = 0
    original_len = len(ims)
    new_intervals = []

    while original_pos < original_len:
        #print()
        #print("original pos", original_pos)
        #print("template pos", template_pos)
        #print("interval indnex", interval_index)

        if interval_index < len(intervals) and original_pos == intervals[interval_index]['start']:
            #print("original pos", original_pos, "aligns with start", intervals[interval_index]['start'])
            start, stop, interp = intervals[interval_index]["start"], intervals[interval_index]["stop"], int(intervals[interval_index]["interp"])
            interval_interp = interpolate(ims[start:stop+1], interp, device, model, include_last=False)
            #print("Interp args: ", start, stop, interp)
            #print("Interpolation output: ", interval_interp.shape)

            num_frames_added = interval_interp.shape[0]
            template[template_pos:template_pos+num_frames_added] = interval_interp

            new_intervals.append({"start": template_pos, "stop": template_pos+num_frames_added, 
                            "interp": int(interp)})


            template_pos += (num_frames_added) # point to last frame added
            original_pos = stop
            interval_index += 1
        else:
            #print("inserting image at original pos", original_pos, "into template pos", template_pos)
            template[template_pos] = copy_and_normalize(ims[original_pos])
            original_pos += 1
            template_pos += 1
        #print("")

    del model
    return template, new_intervals

def interpolate(ims: np.ndarray, exp: int, device, model, include_last=True)->np.ndarray:
    
    '''Interpolates the entirety of ims at scale of exp, meaning that 2^exp interpolated
    images are added between every pair of images in ims. new movie will have len(ims) * (2^exp -1 )
    frames
    
    Parameters:
    ims (np.ndarray) - images to be interpolated. Should be of shape (n x r x c)
    exp (int) - the interpolation level.
    device (str) - device on which model should run. Either cuda or cpu. Call load_model() for device
    model (torch.Net) - the RIFE model. Call load_model() for this
    
    Returns
    interpoalted (np.ndarray) -  a uint8 array of images inlcuding the real images with 2^exp interpolated
                                ims between each pair of real images
    '''
    scale = 1
    
    def make_inference(model, I0, I1, n):
        middle = model.inference(I0, I1, scale)
        if n == 1:
            return [middle]
        first_half = make_inference(model, I0, middle, n=n//2)
        second_half = make_inference(model, middle, I1, n=n//2)
        if n%2:
            return [*first_half, middle, *second_half]
        else:
            return [*first_half, *second_half]
    
    def pad_image(img):
        tmp = 32
        ph = ((h - 1) // tmp + 1) * tmp
        pw = ((w - 1) // tmp + 1) * tmp
        padding = (0, pw - w, 0, ph - h)
        return F.pad(img, padding)
    
    def reformat(im):
        im = cv2.normalize(im, None, 0, 255, cv2.NORM_MINMAX, dtype=cv2.CV_8U)
        return  cv2.merge((im, im, im))
    
    def make_read_buffer(ims):
        read_buffer = Queue(maxsize=500)
        for im in ims:
            read_buffer.put(im)
        return read_buffer
    
    # if user specifies, save last image to add it later
    if include_last:
        final_im = copy.deepcopy(ims[-1])
        final_im = cv2.normalize(final_im, None, 0, 255, cv2.NORM_MINMAX, dtype=cv2.CV_8U)
    
    ims = ims = [reformat(im) for im in ims]
    lastframe = ims[0]
    h,w,_ = lastframe.shape
    ims = ims[1:]
    
    write_buffer = []
    read_buffer = make_read_buffer(ims)

    I1 = torch.from_numpy(np.transpose(lastframe, (2,0,1))).to(device, non_blocking=True).unsqueeze(0).float() / 255.
    I1 = pad_image(I1)
    temp = None # save lastframe when processing static frame
    i = 0
    while True:
        if temp is not None:
            frame = temp
            temp = None
        else:
            try:
                frame = read_buffer.get(False)
            except Empty:
                break
        I0 = I1
        I1 = torch.from_numpy(np.transpose(frame, (2,0,1))).to(device, non_blocking=True).unsqueeze(0).float() / 255.
        I1 = pad_image(I1)
        I0_small = F.interpolate(I0, (32, 32), mode='bilinear', align_corners=False)
        I1_small = F.interpolate(I1, (32, 32), mode='bilinear', align_corners=False)
        ssim = ssim_matlab(I0_small[:, :3], I1_small[:, :3])

        break_flag = False
        if ssim > 0.996:    
            try:    
                frame = read_buffer.get(False) # read a new frame
                temp = frame
            except Empty:
                break_flag = True
                frame = lastframe

            I1 = torch.from_numpy(np.transpose(frame, (2,0,1))).to(device, non_blocking=True).unsqueeze(0).float() / 255.
            I1 = pad_image(I1)
            I1 = model.inference(I0, I1, scale)
            I1_small = F.interpolate(I1, (32, 32), mode='bilinear', align_corners=False)
            ssim = ssim_matlab(I0_small[:, :3], I1_small[:, :3])
            frame = (I1[0] * 255).byte().cpu().numpy().transpose(1, 2, 0)[:h, :w]
        
        if ssim < 0.2:
            output = []
            for i in range((2 ** exp) - 1):
                output.append(I0)

        else:
            output = make_inference(model, I0, I1, 2**exp-1) if exp else []

        write_buffer.append(lastframe)
        for mid in output:
            mid = (((mid[0] * 255.).byte().cpu().numpy().transpose(1, 2, 0)))
            write_buffer.append(mid[:h, :w])

        lastframe = frame
        if break_flag:
            break
        i+=1
    
    if include_last:
        return np.array([im[:,:,0] for im in write_buffer] + [final_im], dtype = np.uint8)
    else:
        return np.array([im[:,:,0] for im in write_buffer], dtype = np.uint8)


import cv2
import numpy as np
import yeastvision.track.utils as f
from yeastvision.utils import showCellNums
import statistics as stats
import matplotlib.pyplot as plt
from skimage.morphology import thin, skeletonize, opening, dilation, erosion, square
from skimage.measure import regionprops, label
from tqdm import tqdm
from yeastvision.track.cell import getBirthFrame

def get_mating_data(mating, cells):
    mating_tracks = correct_mat_tracks(track_mating(mating))
    return merge(cells, mating_tracks)


def track_mating(mating_masks, visualize = False):
    cell_margin = 20
    numbM = len(mating_masks)
    MATC = [[[] for _ in range(numbM)] for _ in range(2)]
    
    # First loop, for leading cell number one
    print("Tracking mating cells")
    print("\tLoop 1/2")
    for im_no, im in tqdm(enumerate(mating_masks)):
        I2 = im.copy()
        IS6 = np.zeros(I2.shape)
        
        if im_no == 0:
            ccel = 1
            I3 = (I2 == ccel)
            [x_cn,y_cn]=f.get_wind_coord1(I3.astype(np.uint8),cell_margin)
            I3A = I3[np.ix_(y_cn, x_cn)]
        else:
            I3A = MATC[0][im_no-1][np.ix_(y_cn, x_cn)]
            
        I3A = skeletonize((I3A>0).astype(np.uint8), method="zhang")
        
        I2A = I2[np.ix_(y_cn,x_cn)]
        
        if np.sum(I2A) == 0:
                I2A = MATC[0][im_no-1][np.ix_(y_cn, x_cn)]
                
        I3B = np.multiply(I3A.astype(np.uint16), I2A.astype(np.uint16))
        ind = stats.mode(I3B[I3B != 0])
        pix = np.nonzero(I2A==ind)
        pix0 = np.nonzero(I2A!=ind)
        I2A[pix]=1
        I2A[pix0]=0
        
        IS6[np.ix_(y_cn, x_cn)] = I2A
        
        x_cn, y_cn = f.get_wind_coord1(IS6.astype(np.uint8), cell_margin)
        
        I22 = np.zeros_like(I2)
        pix1 = np.nonzero(IS6 == 1)
        I2[pix1] = 0
        
        pix2 = np.unique(I2)
        pix2 = pix2[1:]
        
        for ity in range(0,len(pix2)):
            pix4 = np.nonzero(I2 == pix2[ity])
            I22[pix4] = ity+1
            
        MATC[0][im_no] = IS6
        MATC[1][im_no] = I22

    # Extra loop for python code
    for i in range(len(MATC)):
        for j in range(len(MATC[i])):
            if len(MATC[i][j])==0:  # check if the list is empty
                MATC[i][j] = np.zeros(I2.shape)
    
    # visualzation
    """for i in range(initial_point,len(MATC[0])):
        plt.imshow(MATC[1][i],cmap='jet',aspect='auto')
        plt.title(str(i))
        plt.show()"""
        
    # Correct for spurious segmentation
    fil = np.zeros((1,numbM))
    for it0 in range(0,numbM):
        if sum(sum(MATC[1][it0])) !=0:
            fil[0,it0] = 1
    
    if len(np.argwhere(fil == 0))>0:
        s_seg = np.argwhere(fil == 0)[-1][1]
        
        if s_seg > 0:
            for it1 in range(0,s_seg):
                MATC[1][it1] = np.zeros(IS6.shape)
                
    else:
        s_seg = -1

    # Second loop, for sequential labelling of all other cells
    ATC = np.ones(1,)
    print("Loop 2/2")
    while sum(ATC) != 0:
        
        A1 = np.zeros((1,numbM))
        A4 = 1
        A5 = 1
        rang2 = range(s_seg+1,numbM-1)
        for im_no in rang2:
            
            A2 = opening(MATC[1][im_no], square(3))

            if np.sum(A2) == 0 and A5 != 0:
                A1[0, im_no] = 0
                A5 = 1
            else:
                A1[0, im_no] = 1
                A5 = 0
            
                if im_no == 0:
                    ccel = 1
                    I3 = (MATC[1][im_no] == ccel)
                    x_cn, y_cn = f.get_wind_coord1(I3.astype(np.uint8), cell_margin)
                    I3A = I3[np.ix_(y_cn, x_cn)]              
                elif A1[0, im_no] == 1 and A1[0,im_no-1]==0 and A4 != 0:
                    ccel = min(A2[A2 != 0])
                    I3 = (MATC[1][im_no] == ccel)
                    x_cn, y_cn = f.get_wind_coord1(I3.astype(np.uint8), cell_margin)        
                    I3A = I3[np.ix_(y_cn, x_cn)]
                    A4 = 0
                else:
                    I3A = MATC[0][im_no-1][np.ix_(y_cn, x_cn)]
                    I3A[I3A != np.max(I3A)] = 0
                
                
                I2 = MATC[1][im_no+1]  
                                
                I3A = skeletonize((I3A>0).astype(np.uint8), method="zhang")
                I2A = I2[np.ix_(y_cn, x_cn)]
                
                
                if np.sum(I2A) == 0:
                    I2A = MATC[0][im_no-1][np.ix_(y_cn, x_cn)]
                    
                    pC1 = round(I2A.shape[0]/2)
                    pC2 = round(I2A.shape[1]/2)
                    
                    I2AA = np.zeros(I2A.shape)
                    I2AA[pC1, pC2] = 1
                    
                    I2AA = dilation(I2AA, np.ones((9, 9)))
                    
                    I2AA1 = np.multiply(I2AA,I2A)
                    
                    pixR = np.nonzero(I2A != np.max(I2AA1))
                    I2A[pixR] = 0
                                
                
                I3B = np.multiply(I3A.astype(np.uint16), I2A.astype(np.uint16))
                
               
                if np.sum(I3B) == 0:
                    I2A = MATC[0][im_no - 1][np.ix_(y_cn, x_cn)] 
                    pC1 = round(I2A.shape[0] / 2)
                    pC2 = round(I2A.shape[1] / 2)
                
                    I2AA = np.zeros(I2A.shape)
                    I2AA[pC1, pC2] = 1
                    I2AA = dilation(I2AA, square(9))
                    I2AA1 = np.multiply(I2AA,I2A)
                    pixR = np.nonzero(I2A != np.max(I2AA1))
                    I2A[pixR] = 0
                
                
                I3B = np.multiply(I3A.astype(np.uint16), I2A.astype(np.uint16))
                
                                
                if np.sum(I3B) == 0:
                    I2A = MATC[1][im_no][np.ix_(y_cn, x_cn)]
                    I2A = I2A.astype('float64')
                    
                    pC1 = round(I2A.shape[0] / 2)
                    pC2 = round(I2A.shape[1] / 2)
                
                    I2AA = np.zeros(I2A.shape)
                    I2AA[pC1, pC2] = 1
                    I2AA = dilation(I2AA, square(9))
                    I2AA1 = np.multiply(I2AA,I2A)
                    pixR = np.nonzero(I2A != np.max(I2AA1))
                    I2A[pixR] = 0
                    
                   
                I3B = np.multiply(I3A.astype(np.uint16), I2A.astype(np.uint16))
                
                if np.sum(I3B) == 0:
                    I2A = MATC[1][im_no][np.ix_(y_cn, x_cn)]
                    I3A = skeletonize((I2A>0).astype(np.uint8), method="zhang")
                    I3B = np.multiply(I3A.astype(np.uint16), I2A.astype(np.uint16))
                
                if len(I3B[I3B != 0]) > 0:
                    ind = stats.mode(I3B[I3B != 0])
                else:
                    ind = None
                    
                pix = np.nonzero(I2A==ind)
                pixE = np.nonzero(I2A!=ind)
                I2A[pixE]=0
                                 
                I2A[pix] = np.max(MATC[0][numbM-3]) + 1
                
                IS6B = np.zeros(I2.shape)
                IS6B[np.ix_(y_cn, x_cn)] = I2A
                               
                [x_cn,y_cn]=f.get_wind_coord1(IS6B.astype(np.uint8),cell_margin)
                                
                pixF = np.nonzero(IS6B!=0)
                IS6C = MATC[0][im_no]
                
                IS6C[pixF] = np.max(MATC[0][numbM-3])+1
                MATC[0][im_no] = IS6C
                                
                I2D = MATC[1][im_no]
                _, I2E = cv2.threshold(IS6C.astype(np.uint8), 128, 1, cv2.THRESH_BINARY + cv2.THRESH_OTSU)
                I2E = I2E.astype(bool)
                I2D1 = np.multiply(I2D,I2E)
                I2D1 = erosion(I2D1, square(3))
                pix2 = np.nonzero(I2D1!=0)
                
                if len(I2D[pix2]) > 0:
                    cell_E = stats.mode(I2D[pix2])
                else:
                    cell_E = None
                  
                I2D[I2D == cell_E] = 0           
                I22 = np.zeros(I2D.shape)        
                pix2 = np.unique(I2D)
                pix2 = pix2[1:]
                
                for ity in range(0,len(pix2)):
                    
                    BBB = erosion(I2D == pix2[ity], square(3))
                    
                    if np.mean(BBB) > 0:
                        pix4 = np.nonzero(I2D == pix2[ity])
                        I22[pix4] = ity+1
                    else:
                        print()
                         
                MATC[1][im_no] = I22   
        ATC = np.unique(MATC[1][im_no])
        
    # cell array that contains the fully tracked MatSeg masks
    Matmasks = [[[] for _ in range(im_no+1)] for _ in range(1)]
    for ita in range(im_no+1):
        Matmasks[0][ita] = MATC[0][ita]
    
    # Save the tracks
    #sav_path = r'C:\Users\shrey\OneDrive\Desktop\RA_Orlando_Lab\MAT_2_Py_transation\Shreya_Translated\Saved Data\\'
    #name1 = sav_path + exp_name + '_MatSeg_2_Python_v2'
    #np.save(name1, Matmasks)
    
    # visualization of single labels in tracked MatSeg Masks
    """for iv in range(1, 22):
        for its in range(0, im_no+1):
            #its = im_no
            plt.imshow(Matmasks[0][its] == iv)
            plt.title(str(its+1) + 'c =' + str(iv))
            plt.show()"""
    
    # visualization of full tracked MatSeg Masks
    if visualize:
        for its in range(0, numbM-1):
            plt.imshow(Matmasks[0][its],cmap='jet',aspect='auto')
            showCellNums(Matmasks[0][its])
            plt.title(str(its+1))
            plt.show()
    return np.array(Matmasks, dtype = np.uint16)


def correct_mat_tracks(Matmasks):
    # correction of single labels in tracked MatSeg Masks
    nu = Matmasks[0].shape[0]
    MATrack = [[[] for _ in range(nu)] for _ in range(1)]
    Mrack = [[[] for _ in range(nu)] for _ in range(1)]
    cell_margin = 20
    no_obj = np.max(Matmasks[0, nu-1, :, :])
    
    min_size, avg_size, max_size = f.avg_cell_size(Matmasks, nu, no_obj)
    print("correcting tracks")
    print("\t Loop 1/1")
    for iv in tqdm(range(1, no_obj.astype(int)+1)):
        for its in range(nu):
            canvs  = np.zeros(Matmasks[0][0].shape)
            present_cell = Matmasks[0][its]==iv
            
            if sum(sum(present_cell))==0:
                pass
            else:
                [x_cn,y_cn]=f.get_wind_coord1(present_cell.astype(np.uint8),cell_margin)
                I33 = present_cell[np.ix_(y_cn,x_cn)]
            
                if its == 0:
                    past_cell = Matmasks[0][its]==iv
                elif iv==1 and its!=0:
                    past_cell = Mrack[0][its-1]==iv
                elif its==0 and iv!=1:
                    past_cell = Matmasks[0][its]==iv
                elif its!=0 and iv!=1:
                    past_cell = MATrack[0][its-1]==iv
                
                si1 = sum(sum(present_cell))
                si2 = sum(sum(past_cell))
                
                if si1 <= si2-0.1*si2 and np.average(past_cell) > avg_size - 0.2*avg_size:
                    if iv == 1:
                        Mrack[0][its] = past_cell
                    else:
                        pass
                    [x_cn,y_cn]=f.get_wind_coord1(past_cell.astype(np.uint8),cell_margin)
                    I3A = past_cell[np.ix_(y_cn,x_cn)]
                    
                #we can use other morphological features for adaptive thresholding
                elif si1 >= si2+0.13*si2 and si2!=0 and np.average(past_cell) > avg_size - 0.2*avg_size:
                    if iv == 1:
                        Mrack[0][its] = past_cell
                    else:
                        pass
                    [x_cn,y_cn]=f.get_wind_coord1(past_cell.astype(np.uint8),cell_margin)
                    I3A = past_cell[np.ix_(y_cn,x_cn)]
                    
                elif si2 == 0:
                    I3A = present_cell[np.ix_(y_cn,x_cn)]
                    
                else:
                    Mrack[0][its] = present_cell
                    [x_cn,y_cn]=f.get_wind_coord1(present_cell.astype(np.uint8),cell_margin)
                    I3A = present_cell[np.ix_(y_cn,x_cn)]
                    
                if si1 > si2+0.045*si2 and si2 !=0 and np.average(past_cell) > avg_size - 0.2*avg_size:
                    IA = I3A.astype(np.float32) - past_cell[np.ix_(y_cn,x_cn)].astype(np.float32)
                    IA[IA<0] = 0
                    IA = opening(IA, square(3))
                    I3A1 = opening(I3A-IA, square(3))
                    I3A = I3A1
                    
                canvs1 = np.zeros(Matmasks[0][its].shape)
                canvs1[np.ix_(y_cn, x_cn)] = I3A
                pixo = np.nonzero(canvs1 != 0)
                
                if iv == 1:
                    canvs[pixo] = iv
                    MATrack[0][its] = canvs
                else:
                    canvs2 = MATrack[0][its]
                    canvs2[pixo] = iv
                    MATrack[0][its] = canvs2
            
    # visualization of previously tracked MatSeg Masks
    """for iv in range(1,no_obj.astype(int)+1):
        for its1 in range(nu):
            plt.imshow(Matmasks[0][its1]==iv)
            plt.title(str(its1+1) + ' c = ' + str(iv) + ' Before correction')
            plt.show()
            plt.imshow(MATrack[0][its1]==iv)
            plt.title(str(its1+1) + ' c = ' + str(iv) + ' After correction')
            plt.show()"""
            
    # # Save the corrected masks
    # sav_path = r'C:\Users\shrey\OneDrive\Desktop\RA_Orlando_Lab\MAT_2_Py_transation\Correction Code\Saved Data\\'
    # name1 = sav_path + 'Corrected_MatSeg_2_Python'
    # np.save(name1, MATrack)
    return np.array(MATrack, dtype = np.uint16)

def merge(full_masks, Matmasks, visualize = False):
    #Corrected MATracking masks
    numbM = Matmasks[0].shape[0]
    no_obj = np.max(Matmasks[0, numbM-1, :, :])
    MATC = [[[] for _ in range(numbM)] for _ in range(1)]
    new_tracks = np.zeros_like(Matmasks)
    print("Merging mating cells into cell mask")
    print("\t Loop 1/1")
    for cxell in tqdm(range(1,no_obj.astype(int)+1)):
        for im_no, mask in enumerate(full_masks[0:numbM]):
            if cxell == 1:
                I2 = mask.copy()
            else:
                I2 = MATC[0][im_no]
                
            I3 = Matmasks[0][im_no]
            I3 = (I3==cxell)
            
            if np.sum(I3[:])==0:
                pass
            else:
                s1 = np.sum(np.sum(I3))
                I3 = I3.astype(np.uint16)
                
                I2B1 = np.multiply(I2,I3)
                I2BB = skeletonize((I2B1>0), method="zhang")
                I2B1 = np.multiply(I2B1,I2BB)
                
                objs = np.unique(I2B1).reshape(1, -1)
                
                size = [[] for _ in range(2)]
                for i in objs[0][1:]:
                    s = regionprops(label(I2B1==i))[0].area
                    size[0].append(i)
                    size[1].append(s)
                    
                    
                if (len(size[1])==0 or len(size[1])==1):
                    I2A = np.zeros(I2.shape)
                    I2A[I2==size[0][0]] = 1
                    props_Imat = regionprops(label(I3))[0]
                    props_Ifull = regionprops(label(I2A))[0]
                    m1 = np.abs(props_Imat.extent-props_Ifull.extent)/props_Imat.extent
                    m2 = np.abs(props_Imat.area-props_Ifull.area)/props_Imat.area
                    
                    if len(regionprops(label(I3)))>1:
                        bbox = props_Ifull.bbox
                        IZ = I2A[bbox[0]:bbox[2], bbox[1]:bbox[3]]
                    
                    else:
                        if m1 > 0.01 or m2 > 0.09:
                            bbox = props_Imat.bbox
                            IZ = I3[bbox[0]:bbox[2], bbox[1]:bbox[3]]
                        else:
                            bbox = props_Ifull.bbox
                            IZ = I2A[bbox[0]:bbox[2], bbox[1]:bbox[3]]
                                            
                else:
                    size_ind = np.array(size[0])
                    size_con = np.array(size[1])*100/np.sum(np.array(size[1]))
                    m = size_con >= 10
                    size_con_filt = size_con[m]
                    size_ind_filt = size_ind[m]
                    
                    I2A = np.zeros(I2.shape)
                    for i in size_ind_filt:
                        I2A[I2==i] = 1
                        
                    props_Imat = regionprops(label(I3))[0]
                    props_Ifull = regionprops(label(I2A))[0]
                    m1 = np.abs(props_Imat.extent-props_Ifull.extent)/props_Imat.extent
                    m2 = np.abs(props_Imat.area-props_Ifull.area)/props_Imat.area
                                      
                    if len(regionprops(label(I3)))>1:
                        bbox = props_Ifull.bbox
                        IZ = I2A[bbox[0]:bbox[2], bbox[1]:bbox[3]]
                    
                    else:
                        if m1 > 0.03 or m2 > 0.09:
                            bbox = props_Imat.bbox
                            IZ = I3[bbox[0]:bbox[2], bbox[1]:bbox[3]]
                        else:
                            bbox = props_Ifull.bbox
                            IZ = I2A[bbox[0]:bbox[2], bbox[1]:bbox[3]]
                
                I2 = I2.astype(np.uint16)
                I44 = np.zeros(I2.shape)
                I44[bbox[0]:bbox[2], bbox[1]:bbox[3]] = IZ
                pix = np.nonzero(I44)
                new_num = np.max(I2) + 1
                I2[pix] = new_num
                MATC[0][im_no] = I2
                new_tracks[im_no][pix] = new_num
                
                if visualize:
                    plt.imshow(I2==np.max(I2))
                    plt.title('cxell: ' + str(cxell) + ', im_no: ' + str(im_no))
                    plt.show()
    MATC[0].append(MATC[0][-1])
    new_tracks = list(new_tracks)
    new_tracks.append(new_tracks[-1])
    return np.array(MATC[0], dtype = np.uint16), np.array(new_tracks, dtype = np.uint16)




import shutil
import os 
import itertools
from os.path import join
from PyQt5.QtWidgets import QFileDialog
from PyQt5.QtWidgets import QDialog
from yeastvision.models.utils import MODEL_DIR, CUSTOM_MODEL_TYPES, getBuiltInModelTypes
from yeastvision.parts.dialogs import ComboBoxDialog

def custom_model_type_search(key):
    for model in CUSTOM_MODEL_TYPES:
        for k in CUSTOM_MODEL_TYPES[model]:
            if k==key:
                return model
    return "proSeg"

def loadCustomModel(parent):
    filter = "Archive File (*);; Pytorch File (*.pth)"
    currDir = parent.getCurrImDir() if parent.imLoaded else os.getcwd()
    path,_ = QFileDialog.getOpenFileName(
        parent = parent,
        caption = f'''Select Custom Weights to Load. The name of a pre-existing model architecture should be included in the filename if the model
        is a derivative of one of{",".join(parent.modelTypes)}''',
        directory = currDir,
        filter = filter
    )

    if not path:
        return

    customModelFile = os.path.split(path)[1]
    customModelName = customModelFile.split(".")[0]
    validName = False
    
    modelTypes = getBuiltInModelTypes()
    for modelType in modelTypes:
        
        if modelType.lower() == customModelName.lower():
            parent.showError(f"custom model name must not be one of {','.join(modelTypes)}")            
        
        if modelType.lower() in customModelName.lower():
            validName = True
            break
    if not validName:
        choices = list(itertools.chain.from_iterable(CUSTOM_MODEL_TYPES.values()))
        dialog = ComboBoxDialog(choices,
                                "Model Type: ", 
                                "Specify Custom Model Type. Pick the most applicable",
                                parent=parent)
        if dialog.exec_() == QDialog.Accepted:
            selectedModelType = dialog.getSelection()
            if selectedModelType:
                modelType = custom_model_type_search(selectedModelType)
        else:
            parent.showError("Please specify a model type")
            return
    dst = join(MODEL_DIR, modelType, customModelFile)
    shutil.copy2(path,dst)
    parent.newModels()




from yeastvision.models.utils import is_RGB
import torch
import numpy as np
from PyQt5 import QtCore

class SegmentWorker(QtCore.QObject):
    '''Handles Multithreading'''
    finished = QtCore.pyqtSignal(object, object, object, object, object, object, object, object)
    def __init__(self, modelClass,ims, params, exp_idx, weightPath, modelType, mask_template_i = None):
        super(SegmentWorker,self).__init__()
        self.mc = modelClass
        self.ims = ims
        self.params = params
        self.weight = weightPath
        self.mType = modelType
        self.exp_idx = exp_idx
        self.mask_i = mask_template_i

    def run(self):
        im_sample = self.ims[0]
        if is_RGB(im_sample):
            row, col, d = im_sample.shape
        else:
            row, col = im_sample.shape
        
        newImTemplate = np.zeros((len(self.ims), row, col))
        tStart, tStop = int(self.params["T Start"]), int(self.params["T Stop"])
        with torch.no_grad():
            output = self.mc.run(self.ims[tStart:tStop+1],self.params, self.weight)
        self.finished.emit(output, self.mc, newImTemplate, self.params, self.weight, self.mType, self.exp_idx, self.mask_i)

class TrackWorker(QtCore.QObject):
    finished = QtCore.pyqtSignal(object, object, object)

    def __init__(self, func, cells, z, exp_idx, obj = None):
        super(TrackWorker,self).__init__()
        self.trackfunc = func
        self.z = z
        self.cells = cells
        self.exp_idx = exp_idx
        self.obj = obj
    
    def run(self):
        if self.obj is not None:
            out =  self.trackfunc(self.obj, self.cells)
        else:
            out = self.trackfunc(self.cells, transpose_out=True)
        self.finished.emit(self.z, self.exp_idx, out)

class InterpolationWorker(QtCore.QObject):
    finished = QtCore.pyqtSignal(object, object, object, object, object, object, object)

    def __init__(self, ims, newname, annotations, experiment_index, func, intervals, original_len):
        super(InterpolationWorker, self).__init__()
        self.ims = ims
        self.name = newname
        self.annotations = annotations
        self.exp_idx = experiment_index
        self.intervals = intervals
        self.func = func
        self.original_len = original_len

    def run(self):
        ims, new_intervals = self.func(self.ims, self.intervals)
        self.finished.emit(ims, self.exp_idx, self.name, self.annotations, 
                           self.intervals, self.original_len, new_intervals)

class FiestWorker(QtCore.QObject):
    finished = QtCore.pyqtSignal(object, object, object, object)
    def __init__(self, func, exp_idx, channel_id, lineage):
        super(FiestWorker,self).__init__()
        self.func = func
        self.exp = exp_idx
        self.channel_id = channel_id
        self.lineage = lineage
    
    def run(self):
        self.finished.emit(self.func(), self.exp, self.channel_id, self.lineage)


import yeastvision
from yeastvision.models.utils import MODEL_DIR
from yeastvision.ims.interpolate import RIFE_DIR, RIFE_WEIGHTS_PATH
from tqdm import tqdm
import os
import requests

RIFE_FILENAME = "flownet.pkl"
RIFE_URL = f"https://github.com/berkyalcinkaya/yeastvision/blob/main/weights/{RIFE_FILENAME}?raw=True"
def MODEL_URL(model_name): return f"https://github.com/berkyalcinkaya/yeastvision/blob/main/weights/{model_name}?raw=True"
TEST_MOVIE_URL = "https://github.com/berkyalcinkaya/yeastvision/blob/main/data/sample_movie_1"
TEST_MOVIE_NUM_IMS = 10
TEST_MOVIE_ITEMS = ["channel2_cdc", "channel1_phase", "mask1"]
TEST_MOVIE_IM_FORMAT = "im00x"
TEST_MOVIE_DIR = os.path.join(os.path.dirname(yeastvision.__path__[0]), "data/sample_movie_1")


def install_weight(model_name):
    # Construct the URL for the model weights
    
    # Make the request to download the file
    response = requests.get(MODEL_URL(model_name))
    response.raise_for_status()  # This will raise an exception if the request failed
    
    # Define the file path where the weights will be saved
    file_path = os.path.join(MODEL_DIR, model_name, model_name)
    
    # Write the downloaded data to the file
    with open(file_path, 'wb') as file:
        file.write(response.content)

def install_rife():
    response = requests.get(RIFE_URL)
    response.raise_for_status()  # This will raise an exception if the request failed
    # Define the file path where the weights will be saved
    file_path = os.path.join(RIFE_DIR, RIFE_FILENAME)
    
    # Write the downloaded data to the file
    with open(file_path, 'wb') as file:
        file.write(response.content)

def install_test_ims():
    src = TEST_MOVIE_DIR

    os.makedirs(src, exist_ok=True)

    for i in tqdm(range(TEST_MOVIE_NUM_IMS)):
        for extension in TEST_MOVIE_ITEMS:
            im_name = f"{TEST_MOVIE_IM_FORMAT.replace('x', str(i).zfill(1))}_{extension}.tif"
            im_url = f"{TEST_MOVIE_URL}/{im_name}?raw=True"

            response = requests.get(im_url)
            response.raise_for_status()

            outpath = os.path.join(src, im_name)
            with open(outpath, "wb") as file:
                file.write(response.content)

# def do_install():
#     current_directory = os.getcwd()
#     folder_path = "yeastvision_weights.zip"
#     models = [model for model in os.listdir(MODEL_DIR) if os.path.isdir(join(MODEL_DIR, model)) and model != "__pycache__"]
#     print("Extracting weights for the following models", models)

#     unzip_directory = os.path.join(current_directory, 'unzipped_files')
#     os.makedirs(unzip_directory, exist_ok=True)

#     with zipfile.ZipFile(folder_path, 'r') as zip_ref:
#         zip_ref.extractall(unzip_directory)

#     # Move each file into a directory with the same name
#     for root, dirs, files in os.walk(unzip_directory):
#         for file_name in files:
#             print("found", file_name)

#             if "._" not in file_name:
#                 # Get the name of the file without the extension
#                 file_name_without_extension = os.path.splitext(file_name)[0]
                
#                 if file_name_without_extension == "flownet":
#                     destination_directory = RIFE_DIR
#                     print("\tlocated weights for movie interpolation -- moving to", destination_directory)

#                 else:
#                     destination_name = ""
#                     for model in models:
#                         if model in file_name_without_extension:
#                             destination_name = model
#                             print("\tfound", file_name, "belong to", model)
#                             exit
#                     destination_directory = join(MODEL_DIR, destination_name)

#                 # Move the file into the destination directory
#                 source_file_path = os.path.join(root, file_name)
#                 destination_file_path = os.path.join(destination_directory, file_name)
#                 print("\t", destination_file_path)
#                 shutil.move(source_file_path, destination_file_path)

#     # Remove the unzipped folder
#     shutil.rmtree(unzip_directory)
#     print("FINISHED - removing", unzip_directory)


# if __name__ == "__main__":
#     do_install()



